/* sane - Scanner Access Now Easy.
   Copyright (C) 1996, 1997 David Mosberger-Tang and Andreas Beck
   This file is part of the SANE package.

   SANE is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   SANE is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   You should have received a copy of the GNU General Public License
   along with sane; see the file COPYING.  If not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <memory.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>

#include <sys/types.h>
#include <sys/uio.h>

#include <netinet/in.h>

#include <sanei.h>

int
sanei_printf (int fd, const char *format,...)
{
  struct iovec iov[2];
  int multi, *hlpi;
  va_list argp;
  char *hlpc;

  static char buffer[16 * 1024];	/* XXX non-reentrant---fix this! */
  static int bufpos = 0;

  char lenbuf[10];

  va_start (argp, format);

  while (*format)
    {
      switch (*format++)
	{
	case '%':
	  if (*format == '*')
	    {
	      multi = va_arg (argp, int);
	      format++;
	    }
	  else
	    multi = 1;

	  switch (*format++)
	    {
	    case 'c':
	      hlpc = va_arg (argp, int8_t *);
	      memcpy (buffer + bufpos, hlpc, multi);
	      bufpos += multi;
	      break;

	    case 'd':
	      hlpi = va_arg (argp, int32_t *);
	      while (multi--)
		{
		  bufpos += sprintf (buffer + bufpos, "%d ", *hlpi++);
		}
	      break;

	    case 's':
	      hlpc = va_arg (argp, char *);
	      while (multi--)	/* doesn't make much sense here ... fix ? */
		{
		  if (hlpc == NULL)
		    {
		      bufpos += sprintf (buffer + bufpos, "%d| ", 0);
		    }
		  else
		    {
		      bufpos += sprintf (buffer + bufpos, "%d|%s ",
					 (int) strlen (hlpc) + 1, hlpc);
		    }
		}
	      break;
	    }
	  break;

	case '\n':
	  buffer[bufpos++] = '\n';	/* For readability */
	  sprintf (lenbuf, "%7d\n", bufpos);
	  iov[0].iov_base = lenbuf;
	  iov[0].iov_len = 8;
	  iov[1].iov_base = buffer;
	  iov[1].iov_len = bufpos;
	  writev (fd, iov, 2);
	  bufpos = 0;
	  break;

	case ' ':
	  break;

	default:
	  fprintf (stderr, "sanei_printf: unknown format character `%c'",
		   format[-1]);
	  break;
	}
    }
  return bufpos;
}
