/* sane - Scanner Access Now Easy.
   Copyright (C) 1996, 1997 David Mosberger-Tang and Andreas Beck
   This file is part of the SANE package.

   SANE is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   SANE is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   You should have received a copy of the GNU General Public License
   along with sane; see the file COPYING.  If not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <stdlib.h>

#include <netinet/in.h>

#include <sys/types.h>

#include <sanei.h>

int
sanei_scanf (int fd, const char *format,...)
{
  va_list argp;
  int *hlpi, lenhelp, slenhelp;
  char *hlpc;
  char **hlpp;
  int multi, pointer, count = 0;
  ssize_t nread;

  static char buffer[16 * 1024];	/* XXX non-rentrant---fix this! */
  static int bufpos = 0;
  static int buflen = 0;
  char lenbuf[9] = "      0\n";

  va_start (argp, format);

  while (*format)
    {
      while (buflen == 0)
	{
	  nread = read (fd, lenbuf, 8);
	  if (nread < 0)
	    return -1;
	  if (nread != 8)
	    continue;
	  if (lenbuf[7] != '\n')
	    return -1;		/* lost sync */

	  sscanf (lenbuf, "%d", &buflen);
	  if (buflen > sizeof (buffer) - 1)
	    fprintf (stderr, "sanei_scanf: buffer too small ! (%d)\n", buflen);
	  if (read (fd, buffer, buflen) != buflen)
	    fprintf (stderr, "sanei_scanf: short read\n");
	  buffer[buflen] = '\0';	/* Make sure sscanf doesn't run away */
	  bufpos = 0;
	}
      switch (*format++)
	{
	case '%':
	  multi = 1;		/* Defaults : 1 value and  */
	  pointer = 0;		/* no pointer-type return (automallloc) */
	scan_further:
	  switch (*format++)
	    {
	    case '*':		/* Modifier. Read <arg> times. */
	      multi = va_arg (argp, int);
	      goto scan_further;
	    case '&':		/* Modifier. Return malloced pointer to data valid for %&s only. */
	      pointer = 1;
	      goto scan_further;
	    case 'c':
	      hlpc = va_arg (argp, int8_t *);
	      while (multi--)
		{
		  if (bufpos > buflen - 1)
		    break;
		  *hlpc++ = buffer[bufpos++];
		  count++;
		}
	      break;

	    case 'd':
	      hlpi = va_arg (argp, int32_t *);
	      while (multi--)
		{
		  if (1 != sscanf (buffer + bufpos, "%d %n", hlpi++, &lenhelp))
		    break;
		  bufpos += lenhelp;
		  count++;
		}
	      break;

	    case 's':
	      while (multi--)
		{
		  if (1 != sscanf (buffer + bufpos, "%d|%n", &slenhelp, &lenhelp))
		    break;
		  bufpos += lenhelp;
		  count++;
		  if (pointer)
		    {
		      hlpp = va_arg (argp, char **);
		      if (slenhelp == 0)	/* This is a NULL pointer. */
			{
			  *hlpp++ = NULL;
			  bufpos++;	/* Skip trailing space */
			}
		      else
			{
			  hlpc = *hlpp = malloc (slenhelp);	/* Set pointer to malloced space */
			  while (slenhelp--)
			    *hlpc++ = buffer[bufpos++];
			  *(hlpc - 1) = '\0';	/* replace trailing space with \0 */
			}
		    }
		  else
		    {
		      hlpc = va_arg (argp, char *);
		      if (slenhelp == 0)	/* This is a NULL pointer. */
			{
			  *hlpc++ = '\0';	/* Return as empty string ... */
			  bufpos++;	/* Skip trailing space */
			}
		      else
			{
			  while (slenhelp--)
			    *hlpc++ = buffer[bufpos++];
			  *(hlpc - 1) = '\0';	/* replace trailing space with \0 */
			}
		    }
		}
	      break;
	    }
	  break;

	case '\n':
	  buflen = 0;
	  break;

	case ' ':
	  break;

	default:
	  fprintf (stderr, "sanei_scanf: unknown format characters `%c'\n",
		   format[-1]);
	  break;
	}
    }
  return count;
}
